/*
 * $Id: ActionImpl.java,v 1.1.1.1 2002/07/07 10:27:57 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */
package jp.gr.java_conf.abagames.bulletsmorph;

import jp.gr.java_conf.abagames.bulletml.*;
import java.util.*;
import java.io.StringReader;
import org.w3c.dom.*;

/**
 * Action status.
 *
 * @version $Revision: 1.1.1.1 $
 */
public class ActionImpl {
  static public final int NOT_EXIST = Integer.MIN_VALUE;

  public IActionChoice[] action;
  private int repeat;
  public int pc;
  private int waitCnt;
  private float aimSpeed, mvSpeed;
  private int mvspCnt;
  private float aimDrc, mvDrc;
  private boolean isAim;
  private int mvdrCnt;
  private float prvFireDrc, prvFireSpeed;
  private int acclCnt;
  private float aimMx, aimMy, mvMx, mvMy;

  private ActionImpl parent;
  private BulletImpl bullet;

  private BulletmlManager manager;

  public float[] prms;
  public Vector prmsHistory = new Vector();
  public String fileName;

  public Action actionElm;

  private Random rnd = new Random();

  public ActionImpl(ActionImpl impl) {
    Enumeration elms = impl.prmsHistory.elements();
    while ( elms.hasMoreElements() ) {
      prmsHistory.addElement((float[])elms.nextElement());
    }

    fileName = impl.fileName;
    try {
      actionElm = new Action(impl.actionElm.makeDocument());
    } catch ( Exception e ) {
      e.printStackTrace();
    }
    action = new IActionChoice[impl.action.length];
    for ( int i=0 ; i<impl.action.length ; i++ ) {

      try {

      Document reader = impl.action[i].makeDocument();

      IActionChoice newAction = null, da = impl.action[i];
      if ( da instanceof Repeat ) {
        newAction = new Repeat(reader);
      } else if ( da instanceof Action ) {
        newAction = new Action(reader);
      } else if ( da instanceof ActionRef ) {
        newAction = new ActionRef(reader);
      } else if ( da instanceof Fire ) {
        newAction = new Fire(reader);
      } else if ( da instanceof FireRef ) {
        newAction = new FireRef(reader);
      } else if ( da instanceof ChangeSpeed ) {
        newAction = new ChangeSpeed(reader);
      } else if ( da instanceof ChangeDirection ) {
        newAction = new ChangeDirection(reader);
      } else if ( da instanceof Accel ) {
        newAction = new Accel(reader);
      } else if ( da instanceof Wait ) {
        newAction = new Wait(reader);
      } else if ( da instanceof Vanish ) {
        newAction = new Vanish(reader);
      }
      action[i] = newAction;

      } catch ( Exception e ) {
        e.printStackTrace();
      }
    }
  }

  public ActionImpl(BulletmlManager manager) {
    this.manager = manager;
    pc = NOT_EXIST;
  }

  private void init(BulletImpl bullet) {
    repeat = 1; pc = -1;
    waitCnt = mvspCnt = mvdrCnt = acclCnt = 0;
    mvMx = mvMy = aimMx = aimMy = 0;
    prvFireDrc = 0; prvFireSpeed = 1;
    this.bullet = bullet;
    parent = null;
    prms = null;

    fileName = manager.getFileName();
  }

  public void set(Action action, BulletImpl bullet) {
    actionElm = action;
    this.action = action.getContent();
    init(bullet);
  }

  public void set(IActionChoice[] actions, BulletImpl bullet) {
    actionElm = new Action();
    actionElm.setContent(actions);
    this.action = actions;
    init(bullet);
  }

  public void setParams(float[] prms) {
    this.prms = prms;
  }

  public void setRepeat(int repeat) {
    this.repeat = repeat;
  }

  public void setParent(ActionImpl parent) {
    this.parent = parent;
  }

  public void setMoveStatus(int mvdrCnt, float mvDrc, boolean isAim,
   int mvspCnt, float mvSpeed, int acclCnt, float mvMx, float mvMy) {
    this.mvdrCnt = mvdrCnt; this.mvDrc = mvDrc; this.isAim = isAim;
    this.mvspCnt = mvspCnt; this.mvSpeed = mvSpeed;
    this.acclCnt = acclCnt; this.mvMx = mvMx; this.mvMy = mvMy;
  }

  public void setPrvFireStatus(float pfd, float pfs) {
    prvFireDrc = pfd; prvFireSpeed = pfs;
  }

  public void vanish() {
    if ( parent != null ) {
      ActionImpl pr = parent;
      parent = null;
      pr.vanish();
    }
    pc = NOT_EXIST;
  }

  public final boolean isFinished() {
    return ( pc == NOT_EXIST &&
     mvspCnt <= 0 && mvdrCnt <= 0 && acclCnt <= 0 );
  }

  private final String ABSOLUTE_KEYWORD = "absolute";
  private final String RELATIVE_KEYWORD = "relative";
  private final String SEQUENCE_KEYWORD = "sequence";
  private final String AIM_KEYWORD = "aim";

  public final void move() {
    if ( mvspCnt > 0 ) {
      mvspCnt--;
      bullet.speed += mvSpeed;
    }
    if ( mvdrCnt > 0 ) {
      mvdrCnt--;
      /*if ( mvdrCnt == 0 ) {
        if ( isAim ) {
          bullet.direction = bullet.getAimDeg();
        }
      } else {
        bullet.direction += mvDrc;
      }*/
      bullet.direction += mvDrc;
    }
    if ( acclCnt > 0 ) {
      acclCnt--;
      bullet.mx += mvMx;
      bullet.my += mvMy;
    }

    if ( pc == NOT_EXIST ) return;

    if ( waitCnt > 0 ) {
      waitCnt--;
      return;
    }

    for ( ; ; ) {

    pc++;
    if ( pc >= action.length ) {
      repeat--;
      if ( repeat <= 0 ) {
        pc = NOT_EXIST;
        if ( parent != null ) {
          parent.setMoveStatus(
           mvdrCnt, mvDrc, isAim, mvspCnt, mvSpeed, acclCnt, mvMx, mvMy);
          parent.setPrvFireStatus(prvFireDrc, prvFireSpeed);
          bullet.changeAction(this, parent);
        }
        break;
      } else {
        pc = 0;
      }
    }

    IActionChoice ac = action[pc];
    if ( ac instanceof Repeat ) {
      // Repeat action.
      ActionImpl newAction = manager.getActionImplInstance();
      if ( newAction != null ) {
        Repeat rp = (Repeat)ac;
        newAction.set(
         manager.bulletmlUtil.getActionElm(rp.getActionElm(), fileName), bullet);
        float[] actPrms = manager.bulletmlUtil.getActionParams(rp.getActionElm(), prms);
        if ( actPrms == null ) {
          newAction.setParams(prms);
        } else {
          newAction.setParams(actPrms);
        }
        newAction.setRepeat(manager.bulletmlUtil.getIntValue(rp.getTimes(), prms));
        newAction.setParent(this);
        newAction.setMoveStatus(
         mvdrCnt, mvDrc, isAim, mvspCnt, mvSpeed, acclCnt, mvMx, mvMy);
        newAction.setPrvFireStatus(prvFireDrc, prvFireSpeed);
        bullet.changeAction(this, newAction);

        manager.bulletmlUtil.addActionImpl(newAction, false);

        newAction.move();
        break;
      }
    } else if ( ac instanceof Action || ac instanceof ActionRef ) {
      // Action.
      ActionImpl newAction = manager.getActionImplInstance();
      if ( newAction != null ) {
        IActionElmChoice aec = (IActionElmChoice)ac;
        if ( aec instanceof ActionRef &&
         ((ActionRef)aec).getLabel().startsWith(GPEngine.IMPL_LABEL) ) {
          int ii = new Integer(((ActionRef)aec).getLabel().
           substring(GPEngine.IMPL_LABEL.length())).intValue();
          ActionImpl ai = (ActionImpl)manager.getActionsCatalog().elementAt(ii);
          IActionChoice[] actions = ai.action;
          newAction.set(actions, bullet);
          Vector aph = ai.prmsHistory;
          if ( aph.size() > 0 ) {
            // Use prms history.
            float[] prms = (float[])aph.elementAt(
             Math.abs(rnd.nextInt())%aph.size());
            newAction.setParams(prms);
          }
        } else {
          newAction.set(manager.bulletmlUtil.getActionElm(aec, fileName), bullet);
          float[] actPrms = manager.bulletmlUtil.getActionParams(aec, prms);
          if ( actPrms == null ) {
            newAction.setParams(prms);
          } else {
            newAction.setParams(actPrms);
          }
        }
        newAction.setRepeat(1);
        newAction.setParent(this);
        newAction.setMoveStatus(
         mvdrCnt, mvDrc, isAim, mvspCnt, mvSpeed, acclCnt, mvMx, mvMy);
        newAction.setPrvFireStatus(prvFireDrc, prvFireSpeed);
        bullet.changeAction(this, newAction);

        manager.bulletmlUtil.addActionImpl(newAction, false);

        newAction.move();
        break;
      }
    } else if ( ac instanceof Fire || ac instanceof FireRef ) {
      // Fire action.
      IActionChoice fec = (IActionChoice)ac;
      Fire fire = manager.bulletmlUtil.getFireElm(fec, fileName);
      float[] firePrms = manager.bulletmlUtil.getFireParams(fec, prms);
      BulletImpl bi = manager.getBulletImplInstance();
      if ( bi != null ) {
        bi.clr = bullet.clr+1;
        if ( firePrms == null ) {
          bi.setParams(manager.bulletmlUtil.getBulletParams(fire.getBulletElm(), prms));
        } else {
          bi.setParams(firePrms);
        }
        bi.set(manager.bulletmlUtil.getBulletElm(fire.getBulletElm(), fileName),
         bullet.x, bullet.y, bullet.clr+1, bullet);
        Direction d = fire.getDirection();
        if ( d == null ) {
          d = bi.drcElm;
        }
        float drc;
        if ( d != null ) {
          if ( firePrms  == null ) {
            drc = manager.bulletmlUtil.getFloatValue(d.getContent(), prms);
          } else {
            drc = manager.bulletmlUtil.getFloatValue(d.getContent(), firePrms);
          }
          String type = d.getType();
          if ( type != null ) {
            if ( type.equals(AIM_KEYWORD) )  {
              drc += bullet.getAimDeg();
            } else if ( type.equals(SEQUENCE_KEYWORD) ) {
              drc += prvFireDrc;
            } else if ( type.equals(RELATIVE_KEYWORD) ) {
              drc += bullet.direction;
            }
          }
        } else {
          drc = bullet.getAimDeg();
        }
        bi.direction = prvFireDrc = drc;
        Speed s = fire.getSpeed();
        if ( s == null ) {
          s = bi.spdElm;
        }
        float spd = 1;
        if ( s != null ) {
          if ( firePrms == null ) {
            spd = manager.bulletmlUtil.getFloatValue(s.getContent(), prms);
          } else {
            spd = manager.bulletmlUtil.getFloatValue(s.getContent(), firePrms);
          }
          String type = s.getType();
          if ( type != null && (type.equals(RELATIVE_KEYWORD) || type.equals(SEQUENCE_KEYWORD)) ) {
            spd += prvFireSpeed;
          }
        }
        bi.speed = prvFireSpeed = spd;
      }
    } else if ( ac instanceof ChangeSpeed ) {
      // Change speed action.
      Speed s = ((ChangeSpeed)ac).getSpeed();
      String type = s.getType();
      mvspCnt = manager.bulletmlUtil.getIntValue(((ChangeSpeed)ac).getTerm(), prms);
      if ( type != null && type.equals(SEQUENCE_KEYWORD) ) {
        mvSpeed = manager.bulletmlUtil.getFloatValue(s.getContent(), prms);
        //aimSpeed = bullet.speed + mvSpeed*mvspCnt;
      } else {
        aimSpeed = manager.bulletmlUtil.getFloatValue(s.getContent(), prms);
        if ( type != null && (type.equals(RELATIVE_KEYWORD) || type.equals(SEQUENCE_KEYWORD)) ) {
          aimSpeed += bullet.speed;
        }
        mvSpeed = (aimSpeed - bullet.speed) / mvspCnt;
      }
    } else if ( ac instanceof ChangeDirection ) {
      // Change direction action.
      Direction d = ((ChangeDirection)ac).getDirection();
      String type = d.getType();
      mvdrCnt = manager.bulletmlUtil.getIntValue(((ChangeDirection)ac).getTerm(), prms);
      if ( type != null && type.equals(SEQUENCE_KEYWORD) ) {
        isAim = false;
        mvDrc = manager.bulletmlUtil.getFloatValue(d.getContent(), prms);
      } else {
        aimDrc = manager.bulletmlUtil.getFloatValue(d.getContent(), prms);
        if ( type != null && type.equals(ABSOLUTE_KEYWORD) ) {
          isAim = false;
          mvDrc = (aimDrc - bullet.direction)%360;
        } else if ( type != null && type.equals(RELATIVE_KEYWORD) ) {
          isAim = false;
          aimDrc += bullet.direction;
          mvDrc = (aimDrc - bullet.direction)%360;
        } else {
          isAim = true;
          mvDrc = (aimDrc+bullet.getAimDeg() - bullet.direction)%360;
        }
        if ( mvDrc > 180 ) mvDrc -= 360;
        if ( mvDrc < -180 ) mvDrc += 360;
        mvDrc /= mvdrCnt;
      }
    } else if ( ac instanceof Accel ) {
      // Accel bullet.
      Accel al = (Accel)ac;
      Horizontal hrz = al.getHorizontal();
      acclCnt = manager.bulletmlUtil.getIntValue(al.getTerm(), prms);
      if ( hrz != null ) {
        String type = hrz.getType();
        if ( type != null && type.equals(SEQUENCE_KEYWORD) ) {
          mvMx = manager.bulletmlUtil.getFloatValue(hrz.getContent(), prms);
        } else {
          aimMx = manager.bulletmlUtil.getFloatValue(hrz.getContent(), prms);
          if ( type != null && type.equals(RELATIVE_KEYWORD) ) {
            aimMx += bullet.mx;
          }
          mvMx = (aimMx - bullet.mx) / acclCnt;
        }
      }
      Vertical vtc = al.getVertical();
      if ( vtc != null ) {
        String type = vtc.getType();
        if ( type != null && type.equals(RELATIVE_KEYWORD) ) {
          mvMy = manager.bulletmlUtil.getFloatValue(vtc.getContent(), prms);
        } else {
          aimMy = manager.bulletmlUtil.getFloatValue(vtc.getContent(), prms);
          if ( type != null && (type.equals(RELATIVE_KEYWORD) || type.equals(SEQUENCE_KEYWORD)) ) {
            aimMy += bullet.my;
          }
          mvMy = (aimMy - bullet.my) / acclCnt;
        }
      }
    } else if ( ac instanceof Wait ) {
      // Wait action.
      waitCnt = manager.bulletmlUtil.getIntValue(((Wait)ac).getContent(), prms);
      break;
    } else if ( ac instanceof Vanish ) {
      // Vanish action.
      bullet.vanish();
      break;
    }

    }
  }
}
