/*
 * $Id: BulletmlUtil.java,v 1.1.1.1 2002/07/07 10:27:57 kenta Exp $
 *
 * Copyright 2002 Kenta Cho. All rights reserved.
 */
package jp.gr.java_conf.abagames.bulletsmorph;

import jp.gr.java_conf.abagames.bulletml.*;
import java.util.*;
import java.awt.TextArea;
import java.io.StringReader;

/**
 * Utility class for BulletML.
 *
 * @version $Revision: 1.1.1.1 $
 */
public class BulletmlUtil {
  public Hashtable expressions = new Hashtable();

  private char[] expChr;
  public float rank = 0.5f;


  public Hashtable bullets = new Hashtable();
  public Hashtable actions = new Hashtable();
  public Hashtable fires = new Hashtable();

  private TextArea console;

  public BulletmlUtil() { }

  public BulletmlUtil(BulletmlUtil util) {
    this.expressions = (Hashtable)util.expressions.clone();
    this.rank = rank;

    try {

    Enumeration elms = util.bullets.keys();
    bullets = new Hashtable();
    while ( elms.hasMoreElements() ) {
      String lb = (String)elms.nextElement();
      Bullet bl = (Bullet)util.bullets.get(lb);
      //bullets.put(lb, new Bullet(bl.makeDocument()));
      StringReader reader = new StringReader(bl.makeTextDocument());
      bullets.put(lb, new Bullet(reader));
    }
    elms = util.actions.keys();
    actions = new Hashtable();
    while ( elms.hasMoreElements() ) {
      String lb = (String)elms.nextElement();
      Action ac = (Action)util.actions.get(lb);
      //actions.put(lb, new Action(ac.makeDocument()));
      StringReader reader = new StringReader(ac.makeTextDocument());
      actions.put(lb, new Action(reader));
    }
    elms = util.fires.keys();
    fires = new Hashtable();
    while ( elms.hasMoreElements() ) {
      String lb = (String)elms.nextElement();
      Fire fr = (Fire)util.fires.get(lb);
      //fires.put(lb, new Fire(fr.makeDocument()));
      StringReader reader = new StringReader(fr.makeTextDocument());
      fires.put(lb, new Fire(reader));
    }

    } catch ( Exception e ) {
      e.printStackTrace();
    }
  }

  public int getIntValue(String v, float[] prms) {
    return (int)evalExpression(v, prms);
  }

  public float getFloatValue(String v, float[] prms) {
    return evalExpression(v, prms);
  }

  public void setRank(float rk) {
    rank = rk;
  }

  public float getRank() {
    return rank;
  }

  private void evalFloatValue(Expression ep, int stIdx, int lgt, float sign) {
    if ( expChr[stIdx] == '$' ) {
      String label = new String(expChr, stIdx+1, lgt-1);
      if ( label.equals("rand") ) {
        ep.push(0, Expression.STACK_RAND);
      } else if ( label.equals("rank") ) {
        ep.push(0, Expression.STACK_RANK);
      } else {
        int idx;
        try {
          idx = new Integer(label).intValue()-1;
        } catch ( NumberFormatException e ) {
          ep.push(0, Expression.STACK_NUM);
          return;
        }
        ep.push(0, Expression.STACK_VARIABLE+idx);
      }
    } else {
      try {
        ep.push(new Float(new String(expChr, stIdx, lgt)).floatValue()*sign,
         Expression.STACK_NUM);
      } catch ( NumberFormatException e ) {
        ep.push(0, Expression.STACK_NUM);
      }
    }
  }

  private void evalExpPart(Expression ep, int stIdx, int edIdx) {
    int op[] = new int[] {-1, -1};
    while (expChr[stIdx]=='(' && expChr[edIdx-1]==')') {
      stIdx++;
      edIdx--;
    }
    for ( int i = edIdx-1; i >= stIdx ; i-- ) {
      char c = expChr[i];
      if ( c == ')' ) {
        do {
          i--;
        } while ( expChr[i] != '(' );
      } else if ( op[0] < 0 && (c=='*' || c=='/' || c=='%') ) {
	op[0] = i;
      } else if ( c=='+' || c=='-' ) {
	op[1] = i;
	break;
      }
    }
    if ( op[1] < 0 ) {
      if ( op[0] < 0 ) {
        evalFloatValue(ep, stIdx, edIdx-stIdx, 1);
      } else {
	switch(expChr[op[0]]) {
	case '*':
	  evalExpPart(ep, stIdx, op[0]);
          evalExpPart(ep, op[0]+1, edIdx);
          ep.setOperator(Expression.MULTIPLE);
          break;
	case '/':
	  evalExpPart(ep, stIdx, op[0]);
          evalExpPart(ep, op[0]+1, edIdx);
          ep.setOperator(Expression.DIVISION);
          break;
	case '%':
	  evalExpPart(ep, stIdx, op[0]);
          evalExpPart(ep, op[0]+1, edIdx);
          ep.setOperator(Expression.MODULO);
          break;
	}
      }
    } else {
      if ( op[1] == stIdx ) {
	switch (expChr[op[1]]) {
	case '-':
          evalFloatValue(ep, stIdx+1, edIdx-stIdx-1, -1);
          break;
	case '+':
          evalFloatValue(ep, stIdx+1, edIdx-stIdx-1, 1);
          break;
	}
      } else {
	switch(expChr[op[1]]) {
	case '+':
	  evalExpPart(ep, stIdx, op[1]);
          evalExpPart(ep, op[1]+1, edIdx);
          ep.setOperator(Expression.PLUS);
          break;
	case '-':
	  evalExpPart(ep, stIdx, op[1]);
          evalExpPart(ep, op[1]+1, edIdx);
          ep.setOperator(Expression.MINUS);
          break;
	}
      }
    }
  }

  public float evalExpression(String exp, float[] p) {
    Expression ep = (Expression)expressions.get(exp);
    if ( ep == null ) {
      expChr = new char[exp.length()];
      int ecIdx = 0;
      boolean skip = false;
      StringBuffer buf = new StringBuffer(exp);
      int depth = 0;
      boolean balance = true;
      char ch;
      for( int i=0 ; i<buf.length() ; i++ ) {
        ch = buf.charAt(i);
        switch(ch) {
        case ' ':
        case '\n':
          skip = true;
          break;
        case ')':
          depth--;
          if ( depth < 0 ) balance = false;
          break;
        case '(':
          depth++;
          break;
        }
        if ( skip ) {
          skip = false;
        } else {
          expChr[ecIdx] = ch;
          ecIdx++;
        }
      }
      if ( depth != 0 || !balance ) {
        return 0;
      }
      ep = new Expression(this);
      evalExpPart(ep, 0, ecIdx);
      expressions.put(exp, ep);
    }
    return ep.calc(p);
  }


  public void setConsole(TextArea csl) {
    console = csl;
  }

  public void clear() {
    bullets.clear();
    actions.clear();
    fires.clear();
    expressions.clear();
  }

  private final String FILE_LABEL_SEP = ":_:";

  public void addBullet(Bullet blt, String fileName) {
    String label = fileName + FILE_LABEL_SEP + blt.getLabel();
    blt.setLabel(label);
    bullets.put(label, blt);
  }

  public void addAction(Action act, String fileName) {
    String label = fileName + FILE_LABEL_SEP + act.getLabel();
    act.setLabel(label);
    actions.put(label, act);
  }

  public void addFire(Fire fre, String fileName) {
    String label = fileName + FILE_LABEL_SEP + fre.getLabel();
    fre.setLabel(label);
    fires.put(label, fre);
  }

  public Bullet getBulletElm(IBulletElmChoice bec, String fileName) {
    if ( bec instanceof BulletRef ) {
      String label = ((BulletRef)bec).getLabel();
      if ( label.indexOf(FILE_LABEL_SEP) == -1 ) {
        label = fileName + FILE_LABEL_SEP + label;
        ((BulletRef)bec).setLabel(label);
      }
      Bullet blt = (Bullet)bullets.get(label);
      if ( blt == null ) {
        if ( console != null ) {
          console.append("unknown bullet label: " + label + "\n");
        } else {
          System.out.println("unknown bullet label: " + label);
        }
      }
      return blt;
    } else if ( bec instanceof Bullet ) {
      return (Bullet)bec;
    }
    return null;
  }

  public float[] getBulletParams(IBulletElmChoice bec, float[] prms) {
    if ( bec instanceof BulletRef ) {
      BulletRef br = (BulletRef)bec;
      float[] prm = new float[br.getParamCount()];
      for ( int i=prm.length-1; i>=0 ; i-- ) {
        prm[i] = getFloatValue(br.getParam(i).getContent(), prms);
      }
      return prm;
    }
    return null;
  }

  public Action getActionElm(IActionElmChoice aec, String fileName) {
    if ( aec instanceof ActionRef ) {
      String label = ((ActionRef)aec).getLabel();
      if ( label.indexOf(FILE_LABEL_SEP) == -1 ) {
        label = fileName + FILE_LABEL_SEP + label;
        ((ActionRef)aec).setLabel(label);
      }
      Action act = (Action)actions.get(label);
      if ( act == null ) {
        if ( console != null ) {
          console.append("unknown action label: " + label + "\n");
        } else {
          System.out.println("unknown action label: " + label);
        }
      }
      return act;
    } else if ( aec instanceof Action ) {
      return (Action)aec;
    }
    return null;
  }

  public float[] getActionParams(IActionElmChoice aec, float[] prms) {
    if ( aec instanceof ActionRef ) {
      ActionRef ar = (ActionRef)aec;
      float[] prm = new float[ar.getParamCount()];
      for ( int i=prm.length-1; i>=0 ; i-- ) {
        prm[i] = getFloatValue(ar.getParam(i).getContent(), prms);
      }
      return prm;
    }
    return null;
  }

  public Fire getFireElm(IActionChoice fec, String fileName) {
    if ( fec instanceof FireRef ) {
      String label = ((FireRef)fec).getLabel();
      if ( label.indexOf(FILE_LABEL_SEP) == -1 ) {
        label = fileName + FILE_LABEL_SEP + label;
        ((FireRef)fec).setLabel(label);
      }
      Fire fire = (Fire)fires.get(label);
      if ( fire == null ) {
        if ( console != null ) {
          console.append("unknown fire label: " + label + "\n");
        } else {
          System.out.println("unknown fire label: " + label);
        }
      }
      return fire;
    } else if ( fec instanceof Fire ) {
      return (Fire)fec;
    }
    return null;
  }

  public float[] getFireParams(IActionChoice fec, float[] prms) {
    if ( fec instanceof FireRef ) {
      FireRef fr = (FireRef)fec;
      float[] prm = new float[fr.getParamCount()];
      for ( int i=prm.length-1; i>=0 ; i-- ) {
        prm[i] = getFloatValue(fr.getParam(i).getContent(), prms);
      }
      return prm;
    }
    return null;
  }

  private Vector actionImpls;
  private boolean setTopBullet = true;

  public void setActionImplsVector(Vector aiv) {
    actionImpls = aiv;
  }

  public Vector getActionImplsVector() {
    return actionImpls;
  }

  public void setTopBulletEnable(boolean f) {
    setTopBullet = f;
  }

  public boolean addActionToNormBullet = false;

  public void addActionImpl(ActionImpl act, boolean isTop) {
    if ( actionImpls == null ) return;
    if ( isTop && !setTopBullet ) return;
    IActionChoice[] actc = act.action;
    Enumeration elms = actionImpls.elements();
    while ( elms.hasMoreElements() ) {
      ActionImpl ae = (ActionImpl)elms.nextElement();
      if ( act.actionElm == ae.actionElm ) {
        if ( act.prms != null ) ae.prmsHistory.addElement(act.prms);
        return;
      }
    }
    actionImpls.addElement(act);
    if ( act.prms != null ) act.prmsHistory.addElement(act.prms);
  }
}
