/*
 * $Id: Screen.java,v 1.1.1.1 2002/07/07 10:27:57 kenta Exp $
 *
 * Copyright 2001 Kenta Cho. All rights reserved.
 */
package jp.gr.java_conf.abagames.bulletsmorph;

import java.awt.*;
import java.awt.image.*;
import java.applet.Applet;
import java.util.Random;

/**
 * Draw applet screen.
 *
 * @version $Revision: 1.1.1.1 $
 */
public class Screen {
  private int pbuf[], bpbuf[];
  private int pbuf1[], pbuf2[];
  private int bufBldIdx[];
  private MemoryImageSource imageSource;
  private Image bufImage;

  public static final int HEIGHT = 320,
                          WIDTH  = 320;

  private Random rnd = new Random();

  public Screen() {
    pbuf1 = new int[WIDTH * HEIGHT+1];
    pbuf2 = new int[WIDTH * HEIGHT+1];
    pbuf = pbuf1; bpbuf = pbuf2;

    pbuf1[pbuf1.length-1] = pbuf2[pbuf2.length-1] = 0;
    bufBldIdx = new int[WIDTH*HEIGHT];
    for ( int i = WIDTH*HEIGHT-1 ; i>=0 ; i-- ){
      int y = i/WIDTH;
      int x = i-y*HEIGHT;
      x -= WIDTH/2; y -= HEIGHT/2;
      x += (rnd.nextInt()%3);
      y += (rnd.nextInt()%3);
      x += WIDTH/2; y += HEIGHT/2;
      if ( x>=0 && x<WIDTH && y>=0 && y<HEIGHT ) {
        bufBldIdx[i] = x+y*WIDTH;
      } else {
        bufBldIdx[i] = pbuf1.length-1;
      }
    }
  }

  /**
   * make screen buffer(off screen image)
   */
  public void makeBuffer(Component cmp) {
    imageSource = new MemoryImageSource
      (WIDTH, HEIGHT, pbuf, 0, WIDTH);
    imageSource.setAnimated(true);
    bufImage = cmp.createImage(imageSource);
    grp = cmp.getGraphics();
  }

  /**
   * clear screen and make smoke effect
   */
  public final void clear() {
    /*for ( int i=pbuf.length-2 ; i>=0 ; i-- ){
        pbuf[i] = 0xff000000;
    }*/
    System.arraycopy(pbuf, 0, bpbuf, 0, bufBldIdx.length);
    for ( int i=bufBldIdx.length-1 ; i>=0 ; i-- ){
      int px = bpbuf[bufBldIdx[i]];
      int r = (px&0xff0000)>>16;
      r = (255-((255-r)>>1))<<16;
      int g = (px&0x00ff00)>>8;
      g = (255-((255-g)>>1))<<8;
      int b =  px&0x0000ff;
      b = (255-((255-b)>>1));
      pbuf[i] = 0xff000000 | r | g | b;
    }
  }

  private ColorModel colorModel = ColorModel.getRGBdefault();
  private Graphics grp;

  public final void swap() {
    imageSource.newPixels();
    grp.drawImage(bufImage, 0, 0, null);
  }

  private final int alphaBlend(int a, int b, int rate) {
    int tmp;
    tmp = (b & 0xff0000)>>16;
    int red = ((((a & 0xff0000)>>16) - tmp) * rate >>16) + tmp;
    tmp = (b & 0x00ff00)>>8;
    int green = ((((a & 0x00ff00)>>8 ) - tmp) * rate >>16) + tmp;
    tmp = (b & 0x0000ff);
    int blue = ((((a & 0x0000ff)    ) - tmp) * rate >>16) + tmp;
    return (red<<16) | (green<<8) | blue | 0xff000000;
  }

  /**
   * draw alpha blended line
   */
  public final void drawLine(int x1, int y1, int x2, int y2, int color) {
    if ( x1 < 0 ) {
      if ( x2 < 0 ) return;
      y1 = (y1-y2)*x2/(x2-x1)+y2;
      x1 = 0;
    } else if ( x2 < 0 ) {
      y2 = (y2-y1)*x1/(x1-x2)+y1;
      x2 = 0;
    }
    if ( x1 > WIDTH-1 ) {
      if ( x2 > WIDTH-1 ) return;
      y1 = (y1-y2)*(x2-(WIDTH-1))/(x2-x1)+y2;
      x1 = WIDTH-1;
    } else if ( x2 > WIDTH-1 ) {
      y2 = (y2-y1)*(x1-(WIDTH-1))/(x1-x2)+y1;
      x2 = HEIGHT-1;
    }
    if ( y1 < 1 ) {
      if ( y2 < 1 ) return;
      x1 = (x1-x2)*(y2-1)/(y2-y1)+x2;
      y1 = 1;
    } else if ( y2 < 1 ) {
      x2 = (x2-x1)*(y1-1)/(y1-y2)+x1;
      y2 = 1;
    }
    if ( y1 > HEIGHT-2 ) {
      if ( y2 > HEIGHT-2 ) return;
      x1 = (x1-x2)*(y2-(HEIGHT-2))/(y2-y1)+x2;
      y1 = HEIGHT-2;
    } else if ( y2 > HEIGHT-2 ) {
      x2 = (x2-x1)*(y1-(HEIGHT-2))/(y1-y2)+x1;
      y2 = HEIGHT-2;
    }

    int lx = Math.abs(x2 - x1);
    int ly = Math.abs(y2 - y1);

    if ( lx < ly ) {
      if ( ly == 0 ) ly++;
      int ax = ((x2 - x1)<<16) / ly;
      int ay = ((y2 - y1)>>16) | 1;
      int x  = x1<<16;
      int y  = y1;
      for ( int i=ly-1 ; i>=0 ; i--, x+=ax, y+=ay ){
        int px = x>>16;
        int p1 = y * WIDTH + px;
        //pbuf[p1] = color;
        int p2 = p1 + 1;
        int rx = x & 0xffff;
        pbuf[p1] = alphaBlend(pbuf[p1],color,       rx);
        pbuf[p2] = alphaBlend(pbuf[p2],color,0xffff-rx);
      }
    } else {
      if ( lx == 0 ) lx++;
      int ay = ((y2 - y1)<<16) / lx;
      int ax = ((x2 - x1)>>16) | 1;
      int x  = x1;
      int y  = y1<<16;
      for ( int i=lx-1 ; i>0 ; i--, x+=ax, y+=ay ) {
        int py = y>>16;
        int p1 = py * WIDTH + x;
        //pbuf[p1] = color;
        int p2 = p1 + WIDTH;
        int ry = y & 0xffff;
        pbuf[p1] = alphaBlend(pbuf[p1],color,       ry);
        pbuf[p2] = alphaBlend(pbuf[p2],color,0xffff-ry);
      }
    }
  }

}
